__version__ = "0.3.0"
# __version__ has to be defined in the first line

"""
:mod:`spvcm` --- Spatially correlated variance components models
================================================================
"""

from .both_levels import *
from .upper_level import *
from .lower_level import *
from . import plotting
